/*
 * Keypad.c
 *
 * Created: 12/08/2015 18:27:00
 * Author: mike, mike's projects, shalewyn.com
 */ 

# define F_CPU 8000000UL

#include <avr/io.h>
#include <util/delay.h>

// macros for manipulating bits - makes things easier to read in the source code
#define KYBD_bit(b) (1 << (b))
#define KYBD_bit_get(prt, bit) ((prt) & (bit))
#define KYBD_bit_set(prt, bit) ((prt) |= (bit))
#define KYBD_bit_clear(prt, bit) ((prt) &= ~(bit))
#define KYBD_bit_write(c, prt, bit) (c ? KYBD_bit_set(prt, bit) : KYBD_bit_clear(prt, bit))

// set the port definitions for the VIA interface
#define VIA_PORT_0 PORTC
#define VIA_PORT_1 VIA_PORT_0
#define VIA_PORT_2 VIA_PORT_0
#define VIA_PORT_3 VIA_PORT_0
#define VIA_PORT_4 PORTB
#define VIA_PORT_5 VIA_PORT_4
#define VIA_PORT_6 VIA_PORT_4
#define VIA_PORT_7 VIA_PORT_4
#define VIA_PORT_IRQ PORTB

// set the pin definitions for the VIA interface (including the VIA IRQ)
#define VIA_PIN_0 3
#define VIA_PIN_1 2 
#define VIA_PIN_2 1
#define VIA_PIN_3 0
#define VIA_PIN_4 6
#define VIA_PIN_5 2
#define VIA_PIN_6 1
#define VIA_PIN_7 0
#define VIA_PIN_IRQ 7

// this routine converts a keyboard matrix scan result into an ascii code
// the lower nibble of 'scanresult' holds the row and the upper nibble holds the column
unsigned char getAscii(unsigned char scanresult)
{
	unsigned char asciicode = 0;

	switch (scanresult)
	{
		// column 1 (far left)
		case 0b00010001:	asciicode='0';	break;
		case 0b00100001:	asciicode='4';	break;
		case 0b01000001:	asciicode='8';	break;
		case 0b10000001:	asciicode='C';	break;
		
		// column 2
		case 0b00010010:	asciicode='1';	break;
		case 0b00100010:	asciicode='5';	break;
		case 0b01000010:	asciicode='9';	break;
		case 0b10000010:	asciicode='D';	break;
		
		// column 3
		case 0b00010100:	asciicode='2';	break;
		case 0b00100100:	asciicode='6';	break;
		case 0b01000100:	asciicode='A';	break;
		case 0b10000100:	asciicode='E';	break;
		
		// column 4 (far right)
		case 0b00011000:	asciicode='3';	break;
		case 0b00101000:	asciicode='7';	break;
		case 0b01001000:	asciicode='B';	break;
		case 0b10001000:	asciicode='F';	break;
		
		default:			asciicode = 0;
	}
	return(asciicode);
}


// this routine causes a scan of the keyboard matrix. each row is held high by default
// and so the routine pulses each row in turn low and checks each column per pulse
// the resulting 8 bit number is then returned from the routine. the top nibble contains
// the column that the key corresponds to the the lower nibble is the row. 
unsigned char scanKbd(void)
{
	// ensure that 
	PORTD = 0xff;
	unsigned char scanresult;
	
	scanresult = 0;

	for (int row=0;row<=3;row++)
	{
				
		KYBD_bit_clear(PORTD, KYBD_bit(row));
		for (int col=4;col<=7;col++)
		{
			if (KYBD_bit_get(PIND, KYBD_bit(col))==0)
			{
				KYBD_bit_set(scanresult, KYBD_bit(col));
				KYBD_bit_set(scanresult, KYBD_bit(row));
			}
		}
		KYBD_bit_set(PORTD, KYBD_bit(row));
	}
	return(scanresult);
}

// main routine
int main(void)
{
	unsigned char scanresult = 0;
	
	DDRD = 0b0001111;			// port D, set to 0-3 (output), 4-7 (input) 
	PORTD = 0xff;				// port D, pins 0-3 output 1111, pull ups enabled for pins 4-7
	DDRC = 0xff;				// port C, set to input
	PORTC = 0;					// port C, pull ups disabled
	DDRB = 0xff;				// port B, set to input
	PORTB = 0;					// port B, pull ups disabled

	// AVR main loop
	unsigned char asciicode = 0;
    while(1)
    {
		scanresult = scanKbd();			// do a scan of the keyboard matrix
		if (scanresult>=1)				// if a key was pressed then execute the following...
		{
			asciicode = getAscii(scanresult);	// convert the scan result to ascii
			// now set the VIA interface lines (bits 0 - 7) so that it reflects the ascii code
			KYBD_bit_write(KYBD_bit_get(asciicode, KYBD_bit(0)), VIA_PORT_0, KYBD_bit(VIA_PIN_0));
			KYBD_bit_write(KYBD_bit_get(asciicode, KYBD_bit(1)), VIA_PORT_1, KYBD_bit(VIA_PIN_1));
			KYBD_bit_write(KYBD_bit_get(asciicode, KYBD_bit(2)), VIA_PORT_2, KYBD_bit(VIA_PIN_2));
			KYBD_bit_write(KYBD_bit_get(asciicode, KYBD_bit(3)), VIA_PORT_3, KYBD_bit(VIA_PIN_3));
			KYBD_bit_write(KYBD_bit_get(asciicode, KYBD_bit(4)), VIA_PORT_4, KYBD_bit(VIA_PIN_4));
			KYBD_bit_write(KYBD_bit_get(asciicode, KYBD_bit(5)), VIA_PORT_5, KYBD_bit(VIA_PIN_5));
			KYBD_bit_write(KYBD_bit_get(asciicode, KYBD_bit(6)), VIA_PORT_6, KYBD_bit(VIA_PIN_6));
			KYBD_bit_write(KYBD_bit_get(asciicode, KYBD_bit(7)), VIA_PORT_7, KYBD_bit(VIA_PIN_7));
			
			// pulse the VIA IRQ low for 10 milliseconds
 			KYBD_bit_clear(VIA_PORT_IRQ, KYBD_bit(VIA_PIN_IRQ));	// low
			_delay_ms(10);
 			KYBD_bit_set(VIA_PORT_IRQ, KYBD_bit(VIA_PIN_IRQ));		// back high
		}
		else
		{
			// ensure that VIA IRQ is kept high if a key wasn't pressed
 			KYBD_bit_set(VIA_PORT_IRQ, KYBD_bit(VIA_PIN_IRQ));
		}
    }
}